#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
Simon Knight (Design)
Eric Matthews (Design)
Steve Cargill (Code)
Mike Montgomery (additional Code)
Dan Malone (Graphic Design)
Farook & Haroon Shamsher / Joi (Title Music)
Richard Joseph (in-game Music & Sound)
*/

typedef struct {
    unsigned int decoded_size;
    unsigned int encoded_size;
    unsigned int tables_size;
    unsigned int nshorts;
    unsigned int nlongs;
    unsigned int bytestomove;
} comp_table_t;


/*

Summary of the compressed data format:


typedef struct {
    word decoded_size;
    word encoded_size;
    word tables_size;
    word nshorts;
    word nlongs;
    word bytestomove;
} comp_table;


 <-   encoded_size   -> <- tables_size ->
|-----encoded data-----|---dictionary----|-tbl-|--decomp code-|
<- ->
 bytestomove

 <-                             decoded_size                              ->
|-------------------------------decoded data--------------------------------|


The image is scanned for up to 1792 common 32-bit words (short entries) and
up to 1792 common top 24 bits of words (long entries). These form the
dictionary, which is stored compressed above the compressed data.


The dictionary consists of t->nshorts + t->nlongs entries. The entries are
represented by:


       00    WW XX YY    -> previous entry + YYXXWW (long)
       00    WW XX YY ZZ -> previous entry + ZZYYXXWW (short)
       01-09             -> n entries consecutive from the last one
       0A-5B             -> previous entry + (n-&0A)
       5C-AD XX          -> previous entry + ((n-&5C)<<8) + XX
       AE-FF XX YY       -> previous entry + ((n-&AE)<<16) + YYXX


Data is compressed two words at a time, and must be decompressed from the top
down. The compressed form starts with a single byte (at the highest memory
location), split into two nibbles. The least-significant nibble represents
the lower-addressed word, the most-significant nibble the higher. Further
bytes are read from decreasing memory addresses, first for the low word, then
the higher, depending on the contents of the nibble. Words encoding forms
are:


                   0  -> 00000000
      WW XX YY ZZ  1  -> WWXXYYZZ
               ZZ 2-8 -> short[((n-2) << 8) + ZZ]
            YY ZZ 9-F -> (long[((n-9) << 8) + ZZ] << 8) + YY


For example, (45 12 34 56 78 51) represents (12345678 short[345])


The encoding scheme is well suited to 32-bit oriented data, particularly ARM
instructions. Before compression, the data would normally be padded out to an
even number of words with zeros. This extra trailing data is harmless for
AIF.


A six-word header precedes the decompression code, giving details of the data
and dictionary sizes. It also gives an indication of when to shuffle the
final compressed data out of the way when decompressing in place.


*/


int inputCount = 0;
void ssput (int i)
{
    if (i >= ' ') putchar (i); else putchar ('.'); // { printf ("[%02X]", i); inputCount+=3; }
    inputCount++;
    if (inputCount > 15)
    {
        inputCount = 0;
        printf ("\n");
    }
}

// #define SEEINPUT
#define SEEOUTPUT
#ifdef SEEINPUT
#define sput(i)    ssput(i)
#else
#define sput(i)
#endif

int main (int argc, char **argv)
{
    FILE *f;
    int i, length;
    unsigned char *Data;
    unsigned char *dest;
    comp_table_t *comp_table = NULL;
    bool verbose = false;

    if (argc < 2)
    {
        fprintf (stderr, "Usage: uncompressAIF inputfile [outputfile]\n");
        return 1;
    }

    f = fopen (argv[1], "rb");
    if (!f)
    {
        fprintf (stderr, "File to be decompressed \"%s\" not found\n", argv[1]);
        return 2;
    }
    fseek (f, 0, SEEK_END);
    length = ftell (f);
    fseek (f, 0, SEEK_SET);

    Data = (unsigned char *)malloc(length);
    if (!Data)
    {
        printf ("No memory for %ul bytes of data\n", length);
        return 3;
    }
    fread (Data, 1,length, f);
    fclose (f);


    //    Is this an AIF compressed absolute file?
    bool CanBeDecompressed = false;
    int DecompressionCodeAt = 0;
    if (
        (((*((int *)(Data+0x00)) & 0x0f000000) == 0x0b000000) || (*((int *)(Data+0x00)) == 0xe1a00000)) &&        // Branch to decompression code or NOP
        (((*((int *)(Data+0x04)) & 0x0f000000) == 0x0b000000) || (*((int *)(Data+0x04)) == 0xe1a00000)) &&        // Branch to self relocation code or NOP
        (((*((int *)(Data+0x08)) & 0x0f000000) == 0x0b000000) || (*((int *)(Data+0x08)) == 0xe1a00000)) &&        // Branch to zero init code or NOP

        ((*((int *)(Data+0x0c)) & 0x0f000000) == 0x0b000000) &&        // Branch to image entry
        ((*((int *)(Data+0x10)) & 0x0ffdffff) == 0x0f000011)        // SWI OS_Exit
    )
    {
        if (verbose)
            printf ("File: %s\n", argv[1]);
        if ((*((int *)(Data+0x00)) & 0x0f000000) == 0x01a00000)    // mov r0,r0
        {
            if (verbose)
                printf ("%08X\tNOP\t; Decompress code (not used)\n", *((unsigned int *)(Data+0x00)));
        } else
        {
            if ((*((int *)(Data+0x00)) & 0x0f000000) == 0x0b000000)    // Branch
            {
                if ((*((int *)(Data+0x00)) & 0xf0000000) == 0xf0000000)    // Branch Never
                {
                    if (verbose)
                        printf ("%08X\tBranch Never (already decompressed)\n", *((unsigned int *)(Data+0x00)));
                } else
                {
                    DecompressionCodeAt = (8 + (((*((unsigned int *)(Data+0x00)))<<8)>>6)) & 0x03FFFFFC;
                    if (DecompressionCodeAt < 0x80 || DecompressionCodeAt > length-0x200)
                    {
                        if (verbose)
                            printf ("%08X\tBranch to Decompress code at %08Xh -- out of range\n", *((unsigned int *)(Data+0x00)), DecompressionCodeAt);
                    } else
                    {
                        if (verbose)
                            printf ("%08X\tBranch to Decompress code at %08Xh\n", *((unsigned int *)(Data+0x00)), DecompressionCodeAt);
                        CanBeDecompressed = true;
                    }
                }
            } else
            {
                if (verbose)
                    printf ("%08X\tDecompress code -- unexpected mnemonic (should be BRANCH or NOP)\n", *((unsigned int *)(Data+0x00)));
            }
        }
        if (verbose)
        {
            printf ("%08X\tRelocate\n", *((unsigned int *)(Data+0x04)));
            printf ("%08X\tZero-Initialize\n", *((unsigned int *)(Data+0x08)));
            printf ("%08X\tImage Entry Point or Entry Point offset\n", *((unsigned int *)(Data+0x0c)));
            printf ("%08X\tSWI \"OS_Exit\"\t; End program\n", *((unsigned int *)(Data+0x10)));
            printf ("%8X\tRead_only_area_size\n", *((unsigned int *)(Data+0x14)));
            printf ("%8X\tRead_write_area_size\n", *((unsigned int *)(Data+0x18)));
            printf ("%8X\tDebug_area_size\n", *((unsigned int *)(Data+0x1c)));
            printf ("%8X\tImage_zero_init_size\n", *((unsigned int *)(Data+0x20)));
            printf ("%8X\tImage_debug_type\n", *((unsigned int *)(Data+0x24)));
            printf ("%8X\tImage_base\n", *((unsigned int *)(Data+0x28)));
            printf ("%8X\tMinimum_workspace\n", *((unsigned int *)(Data+0x2c)));
            printf ("%8X\tAddress_Mode\n", *((unsigned int *)(Data+0x30)));
            printf ("%8X\tInmage_Data_Base\n", *((unsigned int *)(Data+0x34)));
            printf ("%8X\tReserved_000038\n", *((unsigned int *)(Data+0x38)));
            printf ("%8X\tReserved_00003c\n", *((unsigned int *)(Data+0x3c)));
        }

        if (CanBeDecompressed && DecompressionCodeAt)
        {
            comp_table = (comp_table_t *)(Data+DecompressionCodeAt-6*4);

            if (verbose)
            {
                printf ("\nDecompression table:\n");
                printf ("Decoded size:\t%8Xh\n", comp_table->decoded_size);
                printf ("Original size:\t%8Xh\n", comp_table->encoded_size);
                printf ("Tables size:\t%8Xh\n", comp_table->tables_size);
                printf ("# shorts:\t%8Xh\n", comp_table->nshorts);
                printf ("# longs:\t%8Xh\n", comp_table->nlongs);
                printf ("Bytes to move:\t%8Xh\n", comp_table->bytestomove);
            }

            if (comp_table->nshorts > 0x700)
            {
                fprintf (stderr, "Error: Number of shorts is too large to be possible!\n");
                CanBeDecompressed = false;
            }
            if (CanBeDecompressed && comp_table->nlongs > 0x700)
            {
                fprintf (stderr, "Error: Number of longs is too large to be possible!\n");
                CanBeDecompressed = false;
            }
            if (CanBeDecompressed && comp_table->encoded_size+comp_table->tables_size+0x80 >= (unsigned int)length)
            {
                fprintf (stderr, "Error: Encoded size is larger than actual file size!\n");
                CanBeDecompressed = false;
            }
        }
    } else
    {
        fprintf (stderr, "File %s is not an compressed AIF absolute file\n", argv[1]);
    }

    if (!CanBeDecompressed)
    {
        fprintf (stderr, "Exiting -- no decompressed file produced\n");
        return 4;
    }


    i = DecompressionCodeAt-24-comp_table->tables_size;    // 0x2c8d0-0x8000-0x13af;

    dest = (unsigned char *)malloc(comp_table->decoded_size+0x80);
    if (!dest)
    {
        fprintf (stderr, "No memory for %ul decoded bytes\n", comp_table->decoded_size+0x80);
        return 5;
    }

    memset (dest, 0, comp_table->decoded_size+0x80);
    memcpy (dest, Data, 0x80);
    *((unsigned int *)dest) = 0xfb000000;    // BLNV xxxx

    int previous;
    int count_a;
    int count;
    int value;
    int LoopAgain;

    unsigned int longs[1792];
    unsigned int shorts[1792];

    for (LoopAgain=0; LoopAgain<2; LoopAgain++)
    {
        previous = -1;
        if (!LoopAgain)
            count_a = comp_table->nshorts;
        else
            count_a = comp_table->nlongs;
        count = 0;
        value = -1;

        while (count < count_a)
        {
            if (Data[i] == 0)
            {
                if (LoopAgain == 0)
                {
                    value = previous+(Data[i+1])+(Data[i+2]<<8)+(Data[i+3]<<16)+(Data[i+4]<<24);
                    longs[count] = value;
                    i += 5;
                } else
                {
                    value = previous+(Data[i+1])+(Data[i+2]<<8)+(Data[i+3]<<16);
                    shorts[count] = value;
                    i += 4;
                }
                previous = value;
                count++;
                continue;
            }
            if (Data[i] < 10)
            {
                previous = value;
                for (int j=0; j<Data[i]; j++)
                {
                    previous++;
                    if (LoopAgain == 0)
                    {
                        longs[count] = previous;
                    } else
                    {
                        shorts[count] = previous;
                    }
                    count++;
                }
                value = previous;
                i++;
                continue;
            }
            if (Data[i] < 0x5c)
            {
                value = previous+Data[i]-10;
                i++;
                if (LoopAgain == 0)
                {
                    longs[count] = value;
                } else
                {
                    shorts[count] = value;
                }
                previous = value;
                count++;
                continue;
            }
            if (Data[i] < 0xae)
            {
                value = previous+((Data[i]-0x5c)<<8)+Data[i+1];
                i+=2;
                if (LoopAgain == 0)
                {
                    longs[count] = value;
                } else
                {
                    shorts[count] = value;
                }
                previous = value;
                count++;
                continue;
            }
            //    0xAE - 0xFF
            value = previous+((Data[i]-0xae)<<16)+Data[i+1]+(Data[i+2]<<8);
            i+=3;
            if (LoopAgain == 0)
            {
                longs[count] = value;
            } else
            {
                shorts[count] = value;
            }
            previous = value;
            count++;
        }

    }

    if (i > DecompressionCodeAt-24)
    {
        fprintf (stderr, "ERROR: Out of sync after decoding compression tables!\n");
        return 6;
    }

    i = DecompressionCodeAt-24-comp_table->tables_size-1;    // 0x2c8d0-0x8000-0x13af-1;

    int R4,R1,R3,R6;

    unsigned int *destp = (unsigned int *)(dest+comp_table->decoded_size+0x80);

    while (i > 0x0084)
    {
        R6 = Data[i]; i--;
        R3 = R6 & 15;
        if (R3 >= 9)
        {
            R1 = Data[i]; i--;
            R4 = longs[((R3-9)<<8)+R1];
            destp-=2;
            *destp = R4;
        } else
        {
            if (R3 > 1)
            {
                R1 = Data[i]; i--;
                R4 = Data[i]; i--;
                R4 = (shorts[((R3-2)<<8)+R1]<<8)+R4;
                destp-=2;
                *destp = R4;
            } else
            {
                if (R3 == 0)
                {
                    destp-=2;
                    *destp = 0;
                } else
                {
                    R4 = Data[i]; i--;
                    R1 = Data[i]; i--;
                    R4 |= (R1<<8);
                    R1 = Data[i]; i--;
                    R4 |= (R1<<16);
                    R1 = Data[i]; i--;
                    R4 |= (R1<<24);
                    destp-=2;
                    *destp = R4;
                }
            }
        }
        //    2nd nibble
        R3 = (R6 >> 4) & 15;
        if (R3 >= 9)
        {
            R1 = Data[i]; i--;
            R4 = longs[((R3-9)<<8)+R1];
            destp[1] = R4;
        } else
        {
            if (R3 > 1)
            {
                R1 = Data[i]; i--;
                R4 = Data[i]; i--;
                R4 = (shorts[((R3-2)<<8)+R1]<<8)+R4;
                destp[1] = R4;
            } else
            {
                if (R3 == 0)
                {
                    destp[1] = 0;
                } else
                {
                    R4 = Data[i]; i--;
                    R1 = Data[i]; i--;
                    R4 |= (R1<<8);
                    R1 = Data[i]; i--;
                    R4 |= (R1<<16);
                    R1 = Data[i]; i--;
                    R4 |= (R1<<24);
                    destp[1] = R4;
                }
            }
        }
    }

    char outputname[256];

    if (argc > 2)
    {
        strcpy (outputname, argv[2]);
    } else
    {
        strcpy (outputname, argv[1]);
        strcat (outputname, ".unpacked");
    }

    f = fopen (outputname, "wb");
    if (!f)
    {
        fprintf (stderr, "Unable to create output file \"%s\"\n", outputname);
        return 7;
    }
    fwrite (dest, 1, comp_table->decoded_size+0x80, f);
    fclose (f);
    if (verbose)
        printf ("\nCreated output \"%s\"\n", outputname);

    return 0;
}

